"""Utilities for jdfile."""


from pathlib import Path

from rich.progress import Progress, SpinnerColumn, TextColumn

from jdfile._config.config import Config
from jdfile.models.file import File
from jdfile.models.project import Project
from jdfile.utils.alerts import logger as log


def build_file_list(files: list[Path], config: Config, project: Project = None) -> list[File]:
    """Build a list of files to process. If a directory is specified, all files in the directory will be processed.

        Args:
            config (Config): Configuration object.
            files (list[Path]): List of files or directories to process.
            project (Project, optional): Project object. Defaults to None.

    Returns:
            list[Path]: List of files to process.

    # for possible_file in files:
    #     if possible_file.is_file() and possible_file.stem not in config["ignored_files"]:
    #         list_of_files.append(File(possible_file, terms))
    #     if possible_file.is_dir():
    #         for f in possible_file.rglob("*"):
    #             depth_of_file = len(f.relative_to(possible_file).parts)
    #             if depth_of_file <= depth and f.is_file() and f.stem not in config["ignored_files"]:
    #                 list_of_files.append(File(f, terms))
    """
    directories = [f for f in files if f.is_dir() and f.exists()]
    for _dir in directories:
        files.remove(_dir)
        for f in _dir.rglob("*"):
            depth_of_file = len(f.relative_to(_dir).parts)
            if depth_of_file <= config.depth and f.is_file():
                files.append(f)

    with Progress(
        SpinnerColumn("simpleDotsScrolling"),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Processing files...", total=None)
    all_files = sorted(
        [
            File(path=f, config=config, project=project)
            for f in files
            if f.is_file() and f.name not in config.ignored_files
        ],
        key=lambda x: x.name,
    )

    if config.ignore_dotfiles:
        return [f for f in all_files if not f.is_dotfile]

    log.trace(f"Found {len(all_files)} files.")
    return all_files



"""Model for the File object."""
import difflib
import re
from datetime import datetime
from pathlib import Path
from typing import Any

import rich.repr

from jdfile._config import Config
from jdfile.models.dates import Date
from jdfile.models.project import Project
from jdfile.utils import alerts
from jdfile.utils.alerts import logger as log
from jdfile.utils.enums import Separator
from jdfile.utils.nltk import find_synonyms
from jdfile.utils.questions import select_folder
from jdfile.utils.strings import (
    insert,
    match_case,
    normalize_separators,
    split_camelcase_words,
    split_words,
    strip_special_chars,
    strip_stopwords,
    transform_case,
)


@rich.repr.auto
class File:
    """Representation for a File object."""

    def __init__(
        self, path: Path = None, config: Config = Config(), project: Project = None
    ) -> None:
        """Initialize the File object.

        Args:
            path (Path, optional): Path to the file. Defaults to None.
            config (Config, optional): Configuration object. Defaults to Config().
            project (Project, optional): Project object. Defaults to None.
        """
        self.config = config
        self.accessed = datetime.fromtimestamp(path.stat().st_atime)
        self.created = datetime.fromtimestamp(path.stat().st_ctime)
        self.modified = datetime.fromtimestamp(path.stat().st_mtime)
        self.name = path.name
        self.parent = path.parent
        self.path = path.expanduser().resolve()
        self.suffixes = self.path.suffixes
        self.stem = str(self.path)[: str(self.path).rfind("".join(self.path.suffixes))].replace(
            f"{str(self.parent)}/", ""
        )
        if re.match(r"^\.", self.stem):
            self.is_dotfile: bool = True
        else:
            self.is_dotfile = False
        self.project = project
        self.skip_file = False

        #### Create new attributes for the file ####
        if self.config.date_format:
            self.date = Date(date_format=config.date_format, string=self.stem, ctime=self.created)
        else:
            self.date = None

        if self.config.clean:
            self.new_stem = self._clean_stem()
            self.new_suffixes = self._clean_suffixes()
        else:
            self.new_stem = self.stem
            if self.date and self.date.found_string:
                self.new_stem = re.sub(re.escape(self.date.found_string), "", self.new_stem)
            if self.date and self.date.reformatted_date:
                self.new_stem = insert(
                    self.new_stem,
                    self.date.reformatted_date,
                    self.config.insert_location,
                    self.config.separator,
                )
            self.new_suffixes = self.suffixes.copy()

        self.new_parent = (
            self._get_new_parent()
            if self.project and self.project.exists and config.organize
            else self.parent
        )

    def __rich_repr__(self) -> rich.repr.RichReprResult:  # pragma: no cover
        """Rich representation of the File object.

        Returns:
            rich.repr.RichReprResult: Rich representation of the File object.
        """
        yield "path", self.path
        yield "name", self.name
        yield "stem", self.stem
        yield "parent", self.parent
        yield "suffixes", self.suffixes
        yield "created", self.created
        yield "modified", self.modified
        yield "accessed", self.accessed
        yield "is_dotfile", self.is_dotfile
        yield "target", self.target

    def _clean_stem(self) -> str:
        """Create a cleaned stem for the file.

        Returns:
            str: Cleaned stem for the file.
        """
        new_stem = self.stem

        if self.date and self.date.found_string:
            new_stem = re.sub(re.escape(self.date.found_string), "", new_stem)

        if self.config.split_words:
            new_stem = split_camelcase_words(new_stem, self.config.match_case)

        if self.config.strip_stopwords:
            new_stem = strip_stopwords(new_stem, self.config.stopwords)

        new_stem = strip_special_chars(new_stem)
        new_stem = transform_case(new_stem, self.config.transform_case)
        new_stem = match_case(new_stem, self.config.match_case)
        new_stem = normalize_separators(new_stem, self.config.separator)
        new_stem = new_stem.strip(" -_.")

        if self.date and self.date.reformatted_date:
            new_stem = insert(
                new_stem,
                self.date.reformatted_date,
                self.config.insert_location,
                self.config.separator,
            )

        if self.is_dotfile and not re.match(r"^\.", new_stem):
            return f".{new_stem}"

        return new_stem

    def _clean_suffixes(self) -> list[str]:
        """Clean suffixes from the new filename.

        Returns:
            list[str]: Cleaned suffixes.
        """
        return [
            ".jpg" if ext == ".jpeg" or ext == ".JPEG" else ext.lower() for ext in self.suffixes
        ]

    def _get_new_parent(self) -> Path:
        """Identify the new parent directory for the file.

        Returns:
            Path: New directory for the file.
        """
        words_in_stem = self.tokenize_stem()

        # If a JD number is specified, stop matching other terms
        if match_from_number := [
            i.path for i in self.project.usable_folders if i.number in words_in_stem
        ]:
            log.trace(f"'{self.name}' matched folder from number")
            return match_from_number[0]

        # Match terms to folders
        possible_folders: dict[str, Any] = {}
        for f in self.project.usable_folders:
            key = str(f.path.relative_to(self.project.path))
            for term in [_t.lower() for _t in f.terms]:
                if term.lower() in words_in_stem and key not in possible_folders:
                    possible_folders[key] = [f, [term]]
                if term.lower() in words_in_stem and key in possible_folders:
                    possible_folders[key][1].append(term)

        if len(possible_folders) == 0:
            log.warning(f"Skipping '{self.name}': No matching folders")
            self.skip_file = True
            return self.parent

        if len(possible_folders) == 1 or self.config.force:
            log.trace(f"'{self.name}': 1 matching folder")
            return list(possible_folders.values())[0][0].path

        if len(possible_folders) > 1:
            log.trace(f"'{self.name}': {len(possible_folders)} matching folders")
            result = select_folder(
                possible_folders=possible_folders,
                project_path=self.project.path,
                filename=self.name,
            )
            if result == "skip":
                self.skip_file = True
                return self.parent

            return Path(result)

        # If no matches are found, skip the file
        self.skip_file = True
        return self.parent

    @property
    def target(self) -> Path:
        """Return the target path for the file.

        Returns:
            Path: Target path for the file.
        """
        return Path(self.new_parent / f"{self.new_stem}{''.join(self.new_suffixes)}")

    def commit(self) -> bool:
        """Commit the changes to the file by writing the new filename to disk.

        Returns:
            bool: True if the file was successfully moved, False otherwise.
        """
        if not self.has_changes() or self.skip_file:
            log.info(f"{self.name} -> No changes")
            return False

        if self.project and self.project.exists:
            try:
                display = ".../" + str(self.target.relative_to(self.project.path.parents[0]))
            except ValueError:  # pragma: no cover
                display = str(self.target)
        else:
            display = self.target.name

        if self.target.exists() and self.config.overwrite_existing and not self.target.is_dir():
            if self.config.dry_run:
                alerts.dryrun(f"{self.name} -> {display}")
                return True

            self.path.rename(self.target)
            log.success(f"{self.name} -> {display}")
            return True

        self.unique_name()
        if self.config.dry_run:
            alerts.dryrun(f"{self.name} -> {display}")
            return True
        self.path.rename(self.target)
        log.success(f"{self.name} -> {display}")
        return True

    def has_changes(self) -> bool:
        """Return whether the file has a change.

        Returns:
            True if the file has a change, False otherwise.
        """
        return self.target != self.path

    def print_diff(self) -> str:
        """Print the diff between the original and new stems.

        Returns:
            str: Diff between the original and new stems.
        """
        output = []
        a = f"{self.stem}{''.join(self.suffixes)}"
        b = f"{self.new_stem}{''.join(self.new_suffixes)}"

        matcher = difflib.SequenceMatcher(None, a, b)

        green = "[green reverse]"
        red = "[red reverse]"
        endgreen = "[/]"
        endred = "[/]"

        for opcode, a0, a1, b0, b1 in matcher.get_opcodes():
            if opcode == "equal":
                output.append(a[a0:a1])
            elif opcode == "insert":
                output.append(f"{green}{b[b0:b1]}{endgreen}")
            elif opcode == "delete":
                output.append(f"{red}{a[a0:a1]}{endred}")
            elif opcode == "replace":
                output.append(f"{green}{b[b0:b1]}{endgreen}")
                output.append(f"{red}{a[a0:a1]}{endred}")

        return "".join(output)

    def tokenize_stem(self) -> list[str]:
        """Tokenize the stem of the file.

        Returns:
            list[str]: List of tokens from the stem, all lowercase.
        """
        string = self.new_stem
        string = split_camelcase_words(string)
        words_in_stem = split_words(string)
        words_in_stem.extend(self.config.cli_terms)
        synonyms = []
        if self.config.use_nltk:  # pragma: no cover
            for _w in words_in_stem:
                synonyms.extend(find_synonyms(_w))
        words_in_stem.extend(synonyms)

        return sorted({x.lower() for x in words_in_stem})

    def unique_name(self) -> None:
        """Append an integer to the end of the new filename if the target path already exists."""
        match self.config.separator:
            case Separator.UNDERSCORE:
                sep = "_"
            case Separator.DASH:
                sep = "-"
            case Separator.SPACE:
                sep = " "
            case Separator.NONE:
                sep = ""
            case _:
                sep = "_"

        ########################################
        i = 1
        original_new_stem = f"{self.new_stem}"
        while self.target.exists():
            self.new_stem = f"{original_new_stem}{sep}{i}"
            i += 1



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jdfile', 'jdfile._config', 'jdfile.models', 'jdfile.utils']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.6.0,<0.7.0',
 'nltk>=3.8.1,<4.0.0',
 'questionary>=1.10.0,<2.0.0',
 'rich>=13.3.1,<14.0.0',
 'shellingham>=1.4.0,<2.0.0',
 'tomli>=2.0.1,<3.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['jdfile = jdfile.cli:app']}

setup_kwargs = {
    'name': 'jdfile',
    'version': '1.1.0',
    'description': 'File Manager for the Johnny Decimal System',
    'long_description': '[![PyPI version](https://badge.fury.io/py/jdfile.svg)](https://badge.fury.io/py/jdfile) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/jdfile) [![Python Code Checker](https://github.com/natelandau/jdfile/actions/workflows/automated-tests.yml/badge.svg)](https://github.com/natelandau/jdfile/actions/workflows/automated-tests.yml) [![codecov](https://codecov.io/gh/natelandau/jdfile/branch/main/graph/badge.svg?token=Y11Z883PMI)](https://codecov.io/gh/natelandau/jdfile)\n\n# jdfile\n\n`jdfile` cleans and normalizes filenames. In addition, if you have directories which follow the [Johnny Decimal](https://johnnydecimal.com), jdfile can move your files into the appropriate directory.\n\n`jdfile` cleans filenames based on your preferences.\n\n-   Remove special characters\n-   Trim multiple separators (`word----word` becomes `word-word`)\n-   Normalize to `lower case`, `upper case`, `sentence case`, or `title case`\n-   Normalize all files to a common word separator (`_`, `-`, ` `)\n-   Enforce lowercase file extensions\n-   Remove common English stopwords\n-   Split `camelCase` words into separate words (`camel Case`)\n-   Parse the filename for a date in many different formats\n-   Remove or reformat the date and add it to the the beginning of the filename\n-   Avoid overwriting files by adding a unique integer when renaming/moving\n-   Clean entire directory trees\n-   Optionally, show previews of changes to be made before commiting\n-   Ignore files listed in a config file\n-   Specify casing for words which should never be changed (ie. `iMac` will never be re-cased)\n\n`jdfile` can organize your files into folders.\n\n-   Move files into directory trees following the [Johnny Decimal](https://johnnydecimal.com) system\n-   Parse files and folder names looking for matching terms\n-   Uses [nltk](https://www.nltk.org) to lookup synonyms to improve matching\n-   Add `.jdfile` files to directories containing a list of words that will match files\n\n### Why build this?\n\nIt\'s nearly impossible to file away documents with normalized names when everyone has a different convention for naming files. On any given day, tons of files are attached to emails or sent via Slack by people who have their won way of naming files. For example:\n\n-   `department 2023 financials and budget 08232002.xlsx`\n-   `some contract Jan7 reviewed NOT FINAL (NL comments) v13.docx`\n-   `John&Jane-meeting-notes.txt`\n-   `Project_mockups(WIP)___sep92022.pdf`\n-   `FIRSTNAMElastname Resume (#1) [companyname].PDF`\n-   `code_to_review.js`\n\nIf you are a person who archives documents there are a number of problems with these files.\n\n-   No self-evident way to organize them into folders\n-   No common patterns to search for\n-   Dates all over the place or nonexistent\n-   No consistent casing\n-   No consistent word separators\n-   Special characters within text\n-   I could go on and on...\n\nAdditionally, even if the filenames were normalized, filing documents manually is a pain.\n\n`jdfile` is created to solve for these problems by providing an easy CLI to normalize the filename and organize it into an appropriate directory on your computer.\n\n## Install\n\njdfile requires Python v3.10 or above\n\n```bash\npip install pip install obsidian-metadata\n```\n\n## Usage\n\nRun `jdfile --help` for usage\n\n### Configuration\n\nTo organize files into folders, a valid [toml](https://toml.io/en/) configuration file is required at `~/.jdfile/jdfile.toml`\n\n```toml\n# The name of the project is used as a command line option.\n# (e.g. --organize=project_name)\n[project_name]\n    # (Required) Path to the folder containing the Johnny Decimal project\n    path = "~/johnnydecimal"\n\n    # An optional date format. If specified, the date will be appended to the filename\n    # See https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes for details on how to specify a date.\n    date_format = "None"\n\n    # Ignores dotfiles (files that start with a period) when cleaning a directory.  true or false\n    ignore_dotfiles = true\n\n    # Files in this list will be skipped.\n    ignored_files = [\'file1.txt\', \'file2.txt\']\n\n    # Force the casing of certain words. Great for acronyms or proper nouns.\n    match_case = ["CEO", "CEOs", "iMac", "iPhone"]\n\n    # Overwrite existing files. true or false. If false, unique integers will be appended to the filename.\n    overwrite_existing = false\n\n    # Separator to use between words. Options: "ignore", "underscore", "space", "dash", "none"\n    separator = "ignore"\n\n    # Split CamelCase words into separate words. true or false\n    split_words = false\n\n    # Optional list of project specific stopwords to be stripped from filenames\n    stopwords = ["stopword1", "stopword2"]\n\n    # Strip stopwords from filenames. true or false\n    strip_stopwords = true\n\n    # Transform case of filenames.\n    # Options: "lower", "upper", "title", "CamelCase", "sentence", "ignore",\n    transform_case = "ignore"\n\n    # Use the nltk wordnet corpus to find synonyms for words in filenames. true or false\n    # Note, this will download a large corpus (~400mb) the first time it is run.\n    use_synonyms = false\n```\n\n### Example usage\n\n```bash\n# Normalize all files in a directory to lowercase, with underscore separators\n$ jdfile --case=lower --separator=underscore /path/to/directory\n\n# Clean all files in a directory and confirm all changes before committing them\n$ jdfile --clean /path/to/directory\n\n# Strip common English stopwords from all files in a directory\n$ jdfile --stopwords /path/to/directory\n\n# Transform a date and add it to the filename\n$ jdfile --date-format="%Y-%m-%d" ./somefile_march 3rd, 2022.txt\n\n# Print a tree representation of a Johnny Decimal project\n$ jdfile --project=[project_name] --tree\n\n# Use the settings of a project in the config file to clean filenames without\n# organizing them into folders\n$ jdfile --project=[project_name] --no-organize path/to/some_file.jpg\n\n# Organize files into a Johnny Decimal project with specified terms with title casing\n$ jdfile ---project=[project_name] --term=term1 --term=term2 path/to/some_file.jpg\n```\n\n### Tips\n\nAdding custom functions to your `.bashrc` or `.zshrc` can save time and ensure your filename preferences are always used.\n\n```bash\n# ~/.bashrc\nif command -v jdfile &>/dev/null; then\n\n    clean() {\n        # DESC:\t Clean filenames using the jdfile package\n        if [[ $1 == "--help" || $1 == "-h" ]]; then\n            jdfile --help\n        else\n            jdfile --sep=space --case=title --confirm "$@"\n        fi\n    }\n\n    wfile() {\n        # DESC:\t File work documents\n        if [[ $1 == "--help" || $1 == "-h" ]]; then\n            jdfile --help\n        else\n            jdfile --project=work "$@"\n        fi\n    }\nfi\n```\n\n## Caveats\n\n`jdfile` is built for my own personal use. YMMV depending on your system and requirements. I make no warranties for any data loss that may result from use. I strongly recommend running in `--dry-run` mode prior to updating files.\n\n## Contributing\n\n### Setup: Once per project\n\nThere are two ways to contribute to this project.\n\n#### 1. Local development\n\n1. Install Python 3.10 and [Poetry](https://python-poetry.org)\n2. Clone this repository. `git clone https://github.com/natelandau/jdfile.git`\n3. Install the Poetry environment with `poetry install`.\n4. Activate your Poetry environment with `poetry shell`.\n5. Install the pre-commit hooks with `pre-commit install --install-hooks`.\n\n#### 2. Containerized development\n\n1. Clone this repository. `git clone https://github.com/natelandau/jdfile.git`\n2. Open the repository in Visual Studio Code\n3. Start the [Dev Container](https://code.visualstudio.com/docs/remote/containers). Run <kbd>Ctrl/⌘</kbd> + <kbd>⇧</kbd> + <kbd>P</kbd> → _Remote-Containers: Reopen in Container_.\n4. Run `poetry env info -p` to find the PATH to the Python interpreter if needed by VSCode.\n\n### Developing\n\n-   This project follows the [Conventional Commits](https://www.conventionalcommits.org/) standard to automate [Semantic Versioning](https://semver.org/) and [Keep A Changelog](https://keepachangelog.com/) with [Commitizen](https://github.com/commitizen-tools/commitizen).\n    -   When you\'re ready to commit changes run `cz c`\n-   Run `poe` from within the development environment to print a list of [Poe the Poet](https://github.com/nat-n/poethepoet) tasks available to run on this project. Common commands:\n    -   `poe lint` runs all linters\n    -   `poe test` runs all tests with Pytest\n-   Run `poetry add {package}` from within the development environment to install a run time dependency and add it to `pyproject.toml` and `poetry.lock`.\n-   Run `poetry remove {package}` from within the development environment to uninstall a run time dependency and remove it from `pyproject.toml` and `poetry.lock`.\n-   Run `poetry update` from within the development environment to upgrade all dependencies to the latest versions allowed by `pyproject.toml`.\n',
    'author': 'Nate Landau',
    'author_email': 'github@natenate.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/natelandau/jdfile',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
